#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package net;

use strict;
use Socket;
use threads::shared;
use Carp;
use MIME::Base64;
use libs::base64;

no strict 'refs';

use constant {
	# Wait For
	WF_NONE => 0,
	WF_NICK => 1,
	WF_CHAN => 2,
	WF_ALL => 3,

	MAXBUFLEN => 510,

# These appear to match the implementations I've seen, but are unspecified in the RFCs.
# They may vary by implementation.
	NICKLEN => 30, # some ircds are different. hyperion is 16.
	IDENTLEN => 10, # Sometimes 8 or 9.
			# hyperion may break this due to it's ident format: [ni]=identhere, like this n=northman
	HOSTLEN => 63, # I think I've seen 64 here before.
	MASKLEN => 30 + 10 + 63 + 2, # 105, or maybe 106. the 2 constant is for !@

	CHANNELLEN => 32 # From 005 reply. hyperion is 30.
};

*callfuncs = \&thread::callfuncs;

*itob64 = \&base64::itob64;
*b64toi = \&base64::b64toi;

my $debug = 0;
my $tkn : shared = !$debug;
my $sjoin = 1;
our $noquit = 0;
my $nickip = 1;
my $sjb64 = 1;
our %agents : shared;
my $nl = "\015\012";
our $ircline = 0;
our $ready : shared = 0;
our @defer_join : shared;
our %defer_mode;
our @queue : shared;
our $remoteserv : shared;
our %servers : shared;
our @servernum : shared;
my $setmode_lock : shared;
my @userkill : shared;
our $UnrealProtocolVersion : shared;

our $valid_nick_re = qr/^[][a-zA-Z`\\\|{}_][][a-zA-Z0-9`\\\|{}_-]*$/;

our @opmodes = ('v', 'h', 'o', 'a', 'q');
our %opmodes = (
	v => 1,
	h => 2,
	o => 4,
	a => 8,
	q => 16
);

# Channel modes with arguments:
our $scm = qr/^[bevhoaqI]$/;

# Channel modes with only one setting:
our $ocm = qr/^[kfLlj]$/;

# Allowed channel modes:
our $acm = qr/^[cfijklmnprstzACGIMKLNOQRSTVu]$/;

my %tkn = (
	PRIVMSG 	=> ['PRIVMSG',	'!'],
	WHOIS 		=> ['WHOIS',	'#'],
	WHOWAS		=> ['WHOWAS',	'$'],
	USER		=> ['USER',	'%'],
	NICK		=> ['NICK',	'&'],
	SERVER		=> ['SERVER',	"\'"],
	LIST		=> ['LIST',	'('],
	TOPIC		=> ['TOPIC',	')'],
	INVITE		=> ['INVITE',	'*'],
	VERSION		=> ['VERSION',	'+'],
	QUIT		=> ['QUIT', 	','],
	SQUIT		=> ['SQUIT',	'-'],
	KILL		=> ['KILL',	'.'],
	INFO		=> ['INFO',	'/'],
	LINKS		=> ['LINKS',	'0'],
	STATS		=> ['STATS',	'2'],
	USERS		=> ['USERS',	'3'],
	ERROR		=> ['ERROR',	'5'],
	AWAY		=> ['AWAY',	'6'],
	CONNECT		=> ['CONNECT',	'7'],
	PING		=> ['PING',	'8'],
	PONG 		=> ['PONG',	'9'],
	OPER		=> ['OPER',	';'],
	PASS		=> ['PASS',	'<'],
	WALLOPS		=> ['WALLOPS',	'='],
	GLOBOPS		=> ['GLOBOPS',	']'],
	TIME		=> ['TIME',	'>'],
	NAMES		=> ['NAMES',	'?'],
	SJOIN		=> ['SJOIN',	'~'],
	NOTICE 		=> ['NOTICE',	'B'],
	JOIN		=> ['JOIN',	'C'],
	PART		=> ['PART',	'D'],
	MODE		=> ['MODE',	'G'],
	KICK		=> ['KICK',	'H'],
	USERHOST	=> ['USERHOST',	'J'],
	SQLINE		=> ['SQLINE',	'c'],
	UNSQLINE	=> ['UNSQLINE',	'd'],
	SVSNICK		=> ['SVSNICK',	'e'],
	SVSNOOP		=> ['SVSNOOP',	'f'],
	SVSKILL		=> ['SVSKILL',	'h'],
	SVS2MODE	=> ['SVS2MODE',	'v'],
	CHGHOST		=> ['CHGHOST', 	'AL'],
	CHGIDENT	=> ['CHGIDENT',	'AZ'],
	NETINFO		=> ['NETINFO',	'AO'],
	TSCTL		=> ['TSCTL',	'AW'],
	SWHOIS		=> ['SWHOIS',	'BA'],
	SVSO		=> ['SVSO',	'BB'],
	# One may note... that although there is a TKL Token
	# it does not appear to always be used.
	# Maybe b/c 2 vs 3 chars, nobody cares.
	TKL		=> ['TKL',	'BD'],
	SHUN		=> ['SHUN',	'BL'],
	SVSJOIN		=> ['SVSJOIN',	'BX'],
	SVSPART		=> ['SVSPART',	'BT'],
	SVSSILENCE	=> ['SVSSILENCE','Bs'],
	SVSWATCH	=> ['SVSWATCH',	'Bw'],
	SVSSNO		=> ['SVSSNO',	'BV'],
	SENDSNO		=> ['SENDSNO',	'Ss'],
	
	EOS		=> ['EOS',	'ES'],
	UMODE2		=> ['UMODE2',	"\|"],

	REHASH		=> ['REHASH',	'O']
);

#event::addhandler('EOS', undef(), undef(), __PACKAGE__, 'net::eos');
event::addhandler('NETINFO', undef(), undef(), __PACKAGE__, 'net::netinfo');
event::addhandler('KILL', undef(), undef(), __PACKAGE__, 'net::killed');
event::addhandler('VERSION', undef(), undef(), __PACKAGE__, 'net::version');
event::addhandler('WHOIS', undef(), undef(), __PACKAGE__, 'net::whois');
event::addhandler('SERVER', undef(), undef(), __PACKAGE__, 'net::handle_server');

sub serv_connect() {
	my $remote = $config::conf{remote};
	my $port = $config::conf{port};

	if ($port =~ /\D/) { $port = getservbyname($port, 'tcp') }
	die "No port" unless $port;
	my $iaddr   = inet_aton($remote)               || die "no host: $remote";
	my $paddr   = sockaddr_in($port, $iaddr);
	my $proto   = getprotobyname('tcp');

	socket(IRC, PF_INET, SOCK_STREAM, $proto)  || die "socket: $!";
	connect(IRC, $paddr)    || die "connect: $!";

	#open IRC, "/home/remote/ircd/perlserv-tmp/sync.txt" or die $!;
	{ my $ofh = select(IRC); $|=1; select($ofh); }

	event::addhandler('SEOS', undef(), undef(), __PACKAGE__, 'net::eos');


	ircsendimm('PROTOCTL '.($tkn ? 'TOKEN ' : '').'NICKv2 UMODE2 TKLEXT VHP'.
		($noquit ? ' NOQUIT' : '').($sjoin ? ' SJOIN SJOIN2 SJ3' : '').
		($nickip ? ' NICKIP' : '').
		($sjb64 ? ' SJB64 NS VL' : ''),
		'PASS :'.$config::conf{pass},
		'SERVER '.$config::conf{local}.' 1 '.$config::conf{numeric}.($sjb64 ? ( ':U*-*-'.$config::conf{numeric}.' ') : ' :').$config::conf{info});

}

sub ircrecv {
	while(my $in = <IRC>) {
		last unless thread::alive();
		$in =~ s/\r|\n//sg;
		$ircline++ unless $in =~ /^(?:8|PING)/;
		$in = $ircline . ' ' . $in;
		print ">> $in\n" if $debug;
		thread::ircrecv($in);
	}
}

sub parse_line($) {
	my ($in) = @_;
	return unless $in;
	my $cmd;
	
	if($in =~ /^(@|:)(\S+) (\S+)/) {
		$cmd = $3;
	}
	elsif ($in =~ /^(\S+)/) {
		$cmd = $1;
	}

	print " -- Processing $ircline\n" if $debug;
	print "cmd: $cmd\n" if $debug;

	my %cmdhash = (
		PING		=>	'PING',
		'8'		=>	'PING',

		EOS		=>	'EOS',
		ES		=>	'EOS',

		SERVER		=>	'SERVER',
		"\'"		=>	'SERVER',

		SQUIT		=>	'SQUIT',
		'-'		=>	'SQUIT',

		NETINFO		=>	'NETINFO',
		AO		=>	'NETINFO',

		PROTOCTL	=>	'PROTOCTL',

		JOIN		=>	'JOIN',
		C		=>	'JOIN',

		PART		=>	'PART',
		D		=>	'PART',

		SJOIN		=>	'SJOIN',
		'~'		=>	'SJOIN',

		MODE		=>	'MODE',
		G		=>	'MODE',

		PRIVMSG		=>	'MESSAGE',
		'!'		=>	'MESSAGE',
		NOTICE		=>	'MESSAGE',
		B		=>	'MESSAGE',

		AWAY		=>	'AWAY',
		'6'		=>	'AWAY',

		NICK		=>	'NICK',
		'&'		=>	'NICK',

		QUIT		=>	'QUIT',
		','		=>	'QUIT',

		KILL		=>	'KILL',
		'.'		=>	'KILL',

		KICK		=>	'KICK',
		H		=>	'KICK',

		CHGHOST		=>	'HOST',
		AL		=>	'HOST',
		SETHOST		=>	'HOST',
		AA		=>	'HOST',
		'302'		=>	'HOST',

		'340'		=>	'USERIP',

		CHGIDENT	=>	'IDENT',
		AZ		=>	'IDENT',
		SETIDENT	=>	'IDENT',
		AD		=>	'IDENT',

		TOPIC		=>	'TOPIC',
		')'		=>	'TOPIC',

		UMODE2		=>	'UMODE',
		'|'		=>	'UMODE',

		TSCTL		=>	'TSCTL',
		AW		=>	'TSCTL',

		VERSION		=>	'VERSION',
		'+'		=>	'VERSION',

		TKL		=>	'TKL',
		BD		=>	'TKL',

		WHOIS		=>	'WHOIS',
		'#'		=>	'WHOIS',

		SENDSNO		=>	'SNOTICE',
		Ss		=>	'SNOTICE',

		SMO		=>	'SNOTICE',
		AU		=>	'SNOTICE',

		GLOBOPS		=>	'GLOBOPS',
		']'		=>	'GLOBOPS',
	);

	sub PING($) {
		my ($event, $src, $dst, @args);
		$_[0] =~ /^(?:8|PING) :(\S+)$/;
		# ($event, $src, $dst, $args)
		return ('PING', undef, undef, WF_NONE, $1);
	}

	sub EOS($) {
		my $event;
		$_[0] =~ /^(@|:)(\S+) (?:EOS|ES)/; # Sometimes there's extra crap on the end?
		my $server;
		if ($1 eq '@') {
			$server = $servernum[b64toi($2)];
		}
		else {
			$server = $2;
		}
		set_server_state($server, $ircline);
		return undef() unless get_server_state($remoteserv);
		if($server eq $remoteserv) { $event = 'SEOS' } else { $event = 'EOS' }
		print "Ok. we had EOS\n";
		return ($event, undef, undef, WF_ALL, $server);
	}

	sub SERVER($) {
		#net::debug($_[0]) if $debug;
		if($_[0] =~ /^(?:SERVER|\') (\S+) (\S+) :(U[0-9]+)-([A-Za-z0-9]+)-([0-9]+) (.*)$/) {
		# SERVER test-tab.surrealchat.net 1 :U2307-FhinXeOoZEmM-200 SurrealChat
		# cmd, servername, hopCount, U<protocol>-<buildflags>-<numeric> infoLine
			$remoteserv = $1;
			set_server_state($1, 0, $5);
			$servernum[$5] = $1;
			return ('SERVER', undef, undef, WF_ALL, undef, $1, $2, $6, $5, $3, $4);
			# src, serverName, numHops, infoLine, serverNumeric, protocolVersion, buildFlags
		}
		elsif($_[0] =~ /^(:|@)(\S+) (?:SERVER|\') (\S+) (\d+) (\d+) :(.*)$/) {
		# @38 SERVER test-hermes.surrealchat.net 2 100 :SurrealChat
		# source, cmd, new server, hopCount, serverNumeric, infoLine
			my ($numeric, $name);
			if ($1 eq '@') {
				$name = $servernum[b64toi($2)];
			}
			else {
				$name = $2;
			}
			set_server_state($3, 0, $5);
			return ('SERVER', undef, undef, WF_ALL, $name, $3, $4, $6, $5);
			# src, serverName, numHops, infoLine, serverNumeric
		}
		if($_[0] =~ /^(?:SERVER|\') (\S+) (\S+) :(.*)$/) {
			$remoteserv = $1;
			set_server_state($1, 0);
			return ('SERVER', undef, undef, WF_ALL, undef, $1, $2, $3);
			# src, serverName, numHops, infoLine
		}
		elsif($_[0] =~ /^:(\S+) (?:SERVER|\') (\S+) (\d+) :(.*)$/) {
			# source, new server, hop count, description
			set_server_state($2, 0);
			return ('SERVER', undef, undef, WF_ALL, $1, $2, $3, $4);
			# src, serverName, numHops, infoLine
		}
	}

	sub SQUIT($) {
		if($_[0] =~ /^(?:SQUIT|-) (\S+) :(.*)$/) {
			set_server_state($1, undef());
			return ('SQUIT', undef, undef, WF_ALL, undef, $1, $2);
		}
		elsif($_[0] =~ /^(:|@)(\S+) (?:SQUIT|-) (\S+) :(.*)$/) {
			my $name;
			if ($1 eq '@') {
				$name = $servernum[b64toi($2)];
			}
			else {
				$name = $2;
			}
			set_server_state($3, undef());
			return ('SQUIT', undef, undef, WF_ALL, $name, $3, $4);
		}
	}

	sub NETINFO($) {
		$_[0] =~ /^(?:NETINFO|AO) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) :(.*)$/;
		return ('NETINFO', undef, undef, WF_NONE, $1, $2, $3, $4, $5, $6, $7, $8);
	}

	sub PROTOCTL($) {
		$_[0] =~ /^PROTOCTL (.*)$/;
		return ('PROTOCTL', undef, undef, WF_NONE, $1);
	}

	sub JOIN($) {
		$_[0] =~ /^:(\S+) (?:C|JOIN) (\S+)$/;
		return ('JOIN', undef, 1, WF_CHAN, $1, $2);
	}
	
	sub SJOIN($) {
		if ($_[0] =~ /^(?:\~|SJOIN) (\S+) (\S+) (.*)$/) {
			my ($ts, $cn, $payload) = ($1, $2, $3);
			if ($ts =~ s/^!//) {
				$ts = b64toi($ts);
			}
			return ('SJOIN', undef, undef, WF_CHAN, parse_sjoin($remoteserv, $ts, $cn, $payload));
		}
		elsif($_[0] =~ /^(@|:)(\S+) (?:\~|SJOIN) (\S+) (\S+) (.*)$/) {
			my ($server, $ts, $cn, $payload) = ($2, $3, $4, $5);
			my $server;
			if ($1 eq '@') {
				$server = $servernum[b64toi($2)];
			}
			else {
				$server = $2;
			}
			if ($ts =~ s/^!//) {
				$ts = b64toi($ts);
			}
			return ('SJOIN', undef, undef, WF_CHAN, parse_sjoin($server, $ts, $cn, $payload));
		}
	}

	sub PART($) {
		if($_[0] =~ /^:(\S+) (?:D|PART) (\S+) :(.*)$/) {
			return ('PART', undef, 0, WF_CHAN, $1, $2, $3);
		}
		elsif($_[0] =~ /^:(\S+) (?:D|PART) (\S+)$/) {
			return ('PART', undef, 0, WF_CHAN, $1, $2, undef);
		}
	}

	sub MODE($) {
		$_[0] =~ /^(@|:)(\S+) (?:G|MODE) (\S+) (\S+) (.*)(?: \d+)?$/;
			my $name;
			if ($1 eq '@') {
				$name = $servernum[b64toi($2)];
			}
			else {
				$name = $2;
			}
		return ('MODE', undef, 1, WF_ALL, $name, $3, $4, $5);
	}

	sub MESSAGE($) {
		my ($event, @args);
		if($_[0] =~ /^(@|:)(\S+) (?:\!|PRIVMSG) (\S+) :(.*)$/) {
			my $name;
			if ($1 eq '@') {
				$name = $servernum[b64toi($2)];
			}
			else {
				$name = $2;
			}
			$event = 'PRIVMSG'; @args = ($name, $3, $4);
		}
		elsif($_[0] =~ /^(@|:)(\S+) (?:B|NOTICE) (\S+) :(.*)$/) {
			my $name;
			if ($1 eq '@') {
				$name = $servernum[b64toi($2)];
			}
			else {
				$name = $2;
			}
			$event = 'NOTICE'; @args = ($name, $3, $4);
		}
		$args[1] =~ s/\@${config::conf{local}}.*//io;
		return ($event, 0, 1, WF_ALL, @args);
	}

	sub AWAY($) {
		if($_[0] =~ /^:(\S+) (?:6|AWAY) :(.*)$/) {
			return ('AWAY', undef, undef, WF_ALL, $1, $2);
		}
		elsif($_[0] =~ /^:(\S+) (?:6|AWAY) $/) {
			return ('BACK', undef, undef, WF_ALL, $1);
		}
	}

	sub NICK($) {
		my ($event, @args);
		if($_[0] =~ /^:(\S+) (?:NICK|\&) (\S+) :?(\S+)$/) {
			return ('NICKCHANGE', undef, undef, WF_NICK, $1, $2, $3);
		}
		elsif($_[0] =~ /^(?:NICK|\&) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) :(.*)$/) {
#NICK tab 1 1116196525 northman tabriel.tabris.net test-tab.surrealchat.net 0 +iowghaAxNWzt netadmin.SCnet.ops CgECgw== :Sponsored by Skuld
			my ($nick, $hops, $ts, $ident, $host, $server, $stamp, $modes, $vhost, $IP, $gecos) = 
				($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
			if ($ts =~ s/^!//) {
				$ts = b64toi($ts);
			}
			if ($sjb64 and length($server) <= 2 and $server !~ /\./) {
				$server = $servernum[b64toi($server)];
				
			}
			return ('NICKCONN', undef, undef, WF_NICK, $nick, $hops, $ts, $ident, $host, $server, $stamp, $modes, $vhost, $gecos, 
				join('.', unpack('C4', MIME::Base64::decode($IP)))
			);
		}
		elsif($_[0] =~ /^(?:NICK|\&) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) :(.*)$/) {
#NICK tab 1 1116196525 northman tabriel.tabris.net test-tab.surrealchat.net 0 +iowghaAxNWzt netadmin.SCnet.ops :Sponsored by Skuld
			my ($nick, $hops, $ts, $ident, $host, $server, $stamp, $modes, $vhost, $gecos) =
				($1, $2, $3, $4, $5, $6, $7, $8, $9, $10);
			if ($ts =~ s/^!//) {
				$ts = b64toi($ts);
			}
			if ($sjb64 and length($server) <= 2 and $server !~ /\./) {
				$server = $servernum[b64toi($server)];
				
			}
			return ('NICKCONN', undef, undef, WF_NICK, $nick, $hops, $ts, $ident, $host, $server, $stamp, $modes, $vhost, $gecos);
		}
	}

	sub QUIT($) {
		$_[0] =~ /^:(\S+) (?:QUIT|\,) :(.*)$/;
		return ('QUIT', 0, undef, WF_NICK, $1, $2);
	}

	sub KILL($) {
	#:tabris KILL ProxyBotW :tabris.netadmin.SCnet.ops!tabris (test.)
	#:ProxyBotW!bopm@ircop.SCnet.ops QUIT :Killed (tabris (test.))
		$_[0] =~ /^(@|:)(\S+) (?:KILL|\.) (\S+) :(\S+) \((.*)\)$/;
		my $name;
		if ($1 eq '@') {
			$name = $servernum[b64toi($2)];
		}
		else {
			$name = $2;
		}
		return ('KILL', 0, 1, WF_NICK, $name, $3, $4, $5);
	}

	sub KICK($) {
	#:tabris KICK #diagnostics SurrealBot :i know you don't like this. but it's for science!
		$_[0] =~ /^(@|:)(\S+) (?:KICK|H) (\S+) (\S+) :(.*)$/;
		# source, chan, target, reason
		#$src = 0; #$dst = 2;
		my $name;
		if ($1 eq '@') {
			$name = $servernum[b64toi($2)];
		}
		else {
			$name = $2;
		}
		return ('KICK', 0, undef, WF_CHAN, $name, $3, $4, $5);
	}

	sub HOST($) {
		if($_[0] =~ /^:(\S+) (?:CHGHOST|AL) (\S+) (\S+)$/) {
		#:Agent CHGHOST tabris tabris.netadmin.SCnet.ops
			return ('CHGHOST', 0, 1, WF_CHAN, $1, $2, $3);
			#setter, target, vhost
		}
		elsif($_[0] =~ /^:(\S+) (?:SETHOST|AA) (\S+)$/) {
		#:tabris SETHOST tabris.netadmin.SCnet.ops
			return ('CHGHOST', 0, 1, WF_CHAN, $1, $1, $2);
		}

		elsif ($_[0] =~ /^:(?:\S* )?302 (\S+) :(\S+?)\*?=[+-].*?\@(.*)/) {
		#:serebii.razorville.co.uk 302 leif :Jesture=+~Jesture00@buzz-3F604D09.sympatico.ca
			return ('CHGHOST', 0, 1, WF_CHAN, $1, $2, $3);
		}
	}


	sub USERIP($) {
		$_[0] =~ /^:(?:\S* )?340 (\S+) :(\S+?)\*?=[+-].*?\@((?:\.|\d)*)/;
		return ('USERIP', 0, 1, WF_CHAN, $1, $2, $3);
	}

	sub IDENT($) {
		if($_[0] =~ /^:(\S+) (?:CHGIDENT|AL) (\S+) (\S+)$/) {
			return ('CHGIDENT', 0, 1, WF_ALL, $1, $2, $3);
			#setter, target, IDENT
		}
		elsif($_[0] =~ /^:(\S+) (?:SETIDENT|AD) (\S+)$/) {
			return ('CHGIDENT', 0, 1, WF_ALL, $1, $1, $2);
			#setter, target, ident
		}
	}


	sub TOPIC($) {
		if($_[0] =~ /^(@|:)(\S+) (?:TOPIC|\)) (\S+) (\S+) (\S+) :(.*)$/) {
		#:tabris TOPIC #the_lounge tabris 1089336598 :Small Channel in search of Strong Founder for long term relationship, growth, and great conversation.
		my $name;
		if ($1 eq '@') {
			$name = $servernum[b64toi($2)];
		}
		else {
			$name = $2;
		}
			return ('TOPIC', 0, 1, WF_ALL, $name, $3, $4, $5, $6);
		}
		elsif($_[0] =~ /^(?:TOPIC|\)) (\S+) (\S+) (\S+) :(.*)$/) {
		# src, channel, setter, timestamp, topic
			return ('TOPIC', 0, 1, undef, WF_ALL, $1, $2, $3, $4);
		}
	}

	sub UMODE($) {
	#:tabris | +oghaANWt
		$_[0] =~ /^:(\S+) (?:UMODE2|\|) (\S+)$/;
		# src, umodes
		# a note, not all umodes are passed
		# +s, +O, and +t are not passed. possibly others
		# also not all umodes do we care about.
		# umodes we need care about:
		# oper modes: hoaACN,O oper-only modes: HSq
		# regular modes: rxB,izV (V is only somewhat, as the ircd
		# does the conversions from NOTICE to PRIVSMG for us).

		# Yes, I'm changing the event type on this
		# It's better called UMODE, and easily emulated
		# on IRCds with only MODE.
		return ('UMODE', 0, 0, WF_ALL, $1, $2);
	}
	
	sub WHOIS($) {
	# :tab WHOIS ConnectServ :ConnectServ
		if($_[0] =~ /^:(\S+) (?:WHOIS|\#) (\S+)$/) {
			return ('WHOIS', 0, undef, WF_NONE, $1, $2);
		}
		elsif($_[0] =~ /^:(\S+) (?:WHOIS|\#) (\S+) :(\S+)$/) {
			return ('WHOIS', 0, undef, WF_NONE, $1, $3);
		}
	}

	sub TSCTL($) {
		$_[0] =~ /^:(\S+) (?:TSCTL|AW) alltime$/;
		ircsend(":$config::conf{local} NOTICE $1 *** Server=$config::conf{local} TSTime=".
			time." time()=".time." TSOffset=0");
		return;
	}

	sub VERSION($) {
		$_[0] =~ /^:(\S+) (?:VERSION|\+).*$/;
		return ('VERSION', 0, undef, WF_NONE, $1);
	}

	sub TKL($) {
		if ($_[0] =~ /^(@|:)(\S+) (?:TKL|BD) (.*)$/) {
		# We discard the source anyway.
		#my $server;
		#if ($1 eq '@') {
		#	$server = $servernum[b64toi($2)];
		#}
		#else {
		#	$server = $2;
		#}
			return ('TKL', undef, undef, WF_NONE, parse_tkl("TKL $3"));
		}
		elsif ($_[0] =~ /^(?:TKL|BD) (.*)$/) {
			return ('TKL', undef, undef, WF_NONE, parse_tkl("TKL $1"));
		}
	}

	sub SNOTICE($) {
		$_[0] =~ /^(@|:)(\S+) (SENDSNO|Ss|SMO|AU) ([A-Za-z]) :(.*)$/;
		#@servernumeric Ss snomask :message
		my $name;
		if ($1 eq '@') {
			$name = $servernum[b64toi($2)];
		}
		else {
			$name = $2;
		}
		my $event;
		$event = 'SENDSNO' if(($3 eq 'SENDSNO' or $3 eq 'Ss'));
		$event = 'SMO' if(($3 eq 'SMO' or $3 eq 'AU'));
		return ($event, 0, undef, WF_NONE, $name, $4, $5);
	}

	sub GLOBOPS($) {
		$_[0] =~ /^(@|:)(\S+) (?:GLOBOPS|\]) :(.*)$/;
		#@servernumeric [ :message
		my $name;
		if ($1 eq '@') {
			$name = $servernum[b64toi($2)];
		}
		else {
			$name = $2;
		}
		return ('GLOBOPS', 0, undef, WF_NONE, $name, $3);
	}

	#return unless defined($cmdhash{$cmd});
	unless (defined($cmdhash{$cmd})) {
		#print "Bailing out from $ircline:$cmd for lack of cmdhash\n" if $debug;
		debug($in) if $debug;
		return undef();
	}
	my ($event, $src, $dst, $wf, @args) = &{$cmdhash{$cmd}}($in);
	unless (defined($event)) {
		print "Bailing out from $ircline:$cmd for lack of event\n" if $debug;
		return undef();
	}
	#return unless defined $event;

	my (@recipients, @out);
	if(defined($dst)) {
		#$args[$dst] = lc $args[$dst];
		@recipients = split(/\,/, $args[$dst]);
	}
	#if(defined($src)) { $args[$src] = lc $args[$src]; }

	if(@recipients > 1) {
		foreach my $rcpt (@recipients) {
			$args[$dst] = $rcpt;
			push @out, [$event, $src, $dst, $wf, [@args]];
		}
	} else {
		@out = [$event, $src, $dst, $wf, [@args]];
	}

	return @out;
}

# Helper Functions

sub synced() {
	#return ($sync and $sync < $net::ircline);
	foreach my $s (keys(%net::servers)) {
		my $state = net::get_server_state($s);
		if(defined($state) and (!$state or $state > $net::ircline)) {
			return 0;
		}
	}

	return 1;
}

sub initial_synced() {
	my $state = net::get_server_state($remoteserv);
	if(!$state or $state > $net::ircline) {
		return 0;
	} else {
		return 1;
	}
}

sub set_server_state($$;$) {
	my ($server, $state, $numeric) = @_;

	lock %servers; lock @servernum;
	if($servers{$server}) {
		return if($ircline < $servers{$server}[0]);
		$servers{$server}[0] = $ircline;
		$servers{$server}[1] = $state;
		$servers{$server}[2] = $numeric;
		$servernum[$numeric] = $server;
	} else {
		$servers{$server} = &share( [] );
		$servers{$server}[0] = $ircline;
		$servers{$server}[1] = $state;
		$servers{$server}[2] = $numeric;
		$servernum[$numeric] = $server;
	}
}

sub get_server_state($) {
	my ($server) = @_;

	my $badserver = $config::conf{'unsyncserver'};
	return 1 if($badserver and lc $server eq lc $badserver); # I HATE NEOSTATS
	
	lock %servers;
	if($servers{$server}) {
		return $servers{$server}[1];
	} else {
		return undef;
	}
}

sub handle_server($$$$;$$$) {
# This is mostly a stub function, but we may need the $UnrealProtocolVersion
# at a later date. Plus we may want to maintain a server tree in another module.
	my ($srcServer, $serverName, $numHops, $infoLine, $serverNumeric, $protocolVersion, $buildFlags) = @_;
	$UnrealProtocolVersion = $protocolVersion if defined $protocolVersion;
	#debug("handle_server: $srcServer, $serverName, $numHops, $infoLine, $serverNumeric, $protocolVersion, $buildFlags");
}

sub get_online_servers() {
	my @online_servers;
	foreach my $server (keys(%servers)) {
		push @online_servers, $server if $servers{$server}[1];
	}
	return @online_servers;
}

sub is_agent_in_chan($$) {
	my ($agent, $chan) = @_;
	$agent = lc $agent; $chan = lc $chan;

	lock %agents;

	if($agents{$agent} and $agents{$agent}{CHANS} and $agents{$agent}{CHANS}{$chan}) {
		return 1;
	} else {
		return 0;
	}
}

sub agent_connect($$$$$) {
	my ($nick, $ident, $host, $modes, $gecos) = @_;
	my $time = time();
	if (ref($nick)) {
		($nick, $ident, $host, $modes, $gecos) = @$nick;
	}
	else {
		lock %agents;
	
		$agents{lc $nick} = &share( {} );
		$agents{lc $nick}{CHANS} = &share( {} );
		$agents{lc $nick}{PARMS} = &share( [] );
		for(my $i; $i < @_; $i++) {
			$agents{lc $nick}{PARMS}[$i] = $_[$i];
		}
	}
		
	$host = $config::conf{local} unless $host;
	ircsend($tkn{NICK}[$tkn]." $nick 1 $time $ident $host ".
		($sjb64 ? itob64($config::conf{numeric}) : $config::conf{local}).
		" 1 $modes * :$gecos");

	if(keys(%{$agents{lc $nick}{CHANS}})) {
		foreach my $chan (keys(%{$agents{lc $nick}{CHANS}})) {
			net::ircsend(":$nick ".$tkn{JOIN}[$tkn]." $chan");
			# If we tracked chanmodes for agents, that would go here as well.
		}
	}
}

sub agent_quit($$) {
	my ($nick, $msg) = @_;

	{
		lock %agents;

		delete($agents{lc $nick}{CHANS});
		delete($agents{lc $nick});
	}

	ircsendimm(":$nick ".$tkn{QUIT}[$tkn]." :$msg");
}

sub agent_quit_all($) {
	my ($msg) = @_;
	
	my @agents;
	{
		lock %agents;
		@agents = keys(%agents);
	}

	foreach my $a (@agents) {
		agent_quit($a, $msg);
	}
}

sub is_invalid_agentname($$$) {
	my ($botnick, $botident, $bothost) = @_;

	unless($botnick =~ $valid_nick_re) {
		return "Invalid nickname.";
	}
	unless($botident =~ /^[[:alnum:]_]+$/) {
		return "Invalid ident.";
	}
	unless($bothost =~ /^[[:alnum:].-]+$/) {
		return "Invalid vhost.";
	}
	unless($bothost =~ /\./) {
		return "A vhost must contain at least one dot.";
	}
	return undef;
}

sub join($$) {
	my ($agent, $chan) = @_;

	lock %agents;

	if($agents{lc $agent}) {
		lock %agents;
		$agents{lc $agent}{CHANS}{lc $chan} = 1;
		net::ircsend(":$agent ".$tkn{JOIN}[$tkn]." $chan");
	} else {
		if($ready) {
			debug("Tried to make nonexistent agent ($agent) join channel ($chan)");
		} else {
			print "Deferred join: $agent $chan\n";
			push @defer_join, "$agent $chan";
		}
	}
}

sub part($$$) {
	my ($agent, $chan, $reason) = @_;
	{
		lock %agents;
		delete($agents{lc $agent}{CHANS}{lc $chan});
	}
	net::ircsend(":$agent $tkn{PART}[$tkn] $chan :$reason");
}

# Handler functions

sub pong($$$) {
        my ($src, $cookie, $dst) = @_;
	# This will only make sense if you remember that
	# $src is where it came from, $dst is where it went (us)
	# we're basically bouncing it back, but changing from PING to PONG.
	if (defined($dst) and defined($cookie)) {
		# $dst is always $config::conf{local} anyway...
		# this is only valid b/c we never have messages routed THROUGH us
		# we are always an end point.
		ircsendimm(":$dst $tkn{PONG}[$tkn] $src :$cookie");
	}
	else {
		ircsendimm("$tkn{PONG}[$tkn] :$src");
        }
}

sub eos {
	print "GOT EOS\n\n" if $debug;

	#foreach my $k (keys %servers) {
	#	print "Server: $k ircline: ",$servers{$k}[0], " state: ", $servers{$k}[1], "\n";
	#}
	#print "Synced: ", synced(), "\n\n";
	#exit;
	
	ircsendimm(':'.$config::conf{local}.' '.$tkn{EOS}[$tkn]);
	#$ready = 1;
	foreach my $j (@defer_join) {
		print "Processing join: $j\n";
		my ($agent, $chan) = split(/ /, $j);
		net::join($agent, $chan);
	}
	undef(@defer_join);
	
	ircsendimm(@queue) if @queue;
	$ready = 1;
}

sub netinfo($$$$$$$$) {
	ircsendimm($tkn{NETINFO}[$tkn].' 0 '.time." $_[2] $_[3] 0 0 0 :$_[7]");
	$config::conf{network} = $_[7];
}

sub killed($$$$) {
	my ($src, $dst, $path, $reason) = @_;
	lock %agents;
	if (defined($agents{lc $dst})) {
		if (defined ($agents{lc $dst}{KILLED}) and ($agents{lc $dst}{KILLED} == time())) {
			if ($agents{lc $dst}{KILLCOUNT} > 3) {
				debug("Caught in a kill loop for $dst, dying now.");
				thread::shutdown();
			} else {
				$agents{lc $dst}{KILLCOUNT}++;
			}
		} else {
			$agents{lc $dst}{KILLED} = time();
			$agents{lc $dst}{KILLCOUNT} = 1;
		}

		if($src =~ /\./) {
			irckill($config::conf{'local'}, $dst, "Nick Collision");
		} elsif (defined($agents{lc $src})) {
			# Do Nothing.
		} else {
			irckill($main::rsnick, $src, "Do not kill services agents.");
		}
		
		&agent_connect($agents{lc $dst}{PARMS}) if synced();
	}
}

sub tssync {
	ircsendimm(($sjb64 ? '@'.itob64($config::conf{numeric}) : ':'.$config::conf{local})." $tkn{TSCTL}[$tkn] SVSTIME ".time);
}

sub parse_sjoin($$$$) {
	my ($server, $ts, $cn, $parms) = @_;
	my (@users, @bans, @excepts, @invex, @blobs, $blobs, $chmodes, $chmodeparms);
	
	$server = '' unless $server;

	if($parms =~ /^:(.*)/) {
		$blobs = $1;
	} else {
		($chmodes, $blobs) = split(/ :/, $parms, 2);
		($chmodes, $chmodeparms) = split(/ /, $chmodes, 2);
	}
	@blobs = split(/ /, $blobs);
	
	foreach my $x (@blobs) {
		if($x =~ /^(\&|\"|\')(.*)$/) {
			my $type;
			push @bans, $2 if $1 eq '&';
			push @excepts, $2 if $1 eq '"';
			push @invex, $2 if $1 eq "\'";
		} else {
			$x =~ /^([*~@%+]*)(.*)$/;
			my ($prefixes, $nick) = ($1, $2);
			my @prefixes = split(//, $prefixes);
			my $op;
			foreach my $prefix (@prefixes) {
				$op |= $opmodes{q} if ($prefix eq '*');
				$op |= $opmodes{a} if ($prefix eq '~');
				$op |= $opmodes{o} if ($prefix eq '@');
				$op |= $opmodes{h} if ($prefix eq '%');
				$op |= $opmodes{v} if ($prefix eq '+');
			}

			push @users, { NICK => $nick, __OP => $op };
		}
	}
	
	return ($server, $cn, $ts, $chmodes, $chmodeparms, \@users, \@bans, \@excepts, \@invex);
}

sub whois {
#:wyvern.surrealchat.net 311 blah2 tabris northman SCnet-E5870F84.dsl.klmzmi.ameritech.net * :Sponsored by Skuld
#:wyvern.surrealchat.net 307 blah2 tabris :is a registered nick
#:wyvern.surrealchat.net 312 blah2 tabris wyvern.surrealchat.net :SurrealChat - aphrodite.wcshells.com - Chicago.IL
#:wyvern.surrealchat.net 671 blah2 tabris :is using a Secure Connection
#:wyvern.surrealchat.net 317 blah2 tabris 54 1118217330 :seconds idle, signon time
#:wyvern.surrealchat.net 401 blah2 nikanoru :No such nick/channel
#:wyvern.surrealchat.net 311 blah2 somebot bot SCnet-DA158DBF.hsd1.nh.comcast.net * :Some sort of bot
#:wyvern.surrealchat.net 312 blah2 somebot nascent.surrealchat.net :SurrealChat - Hub
#:wyvern.surrealchat.net 335 blah2 somebot :is a Bot on SurrealChat.net
#:wyvern.surrealchat.net 318 blah2 tabris,nikanoru,somebot :End of /WHOIS list.

# Also reference http://www.alien.net.au/irc/irc2numerics.html

	my ($src, $nicklist) = @_;

	my @nicks = split(/\,/, $nicklist);
	my @reply;
	foreach my $nick (@nicks) {
		if (defined($agents{lc $nick})) {
			my ($nick, $ident, $host, $modes, $gecos) = @{$agents{lc $nick}{PARMS}};
			$host = $config::conf{local} unless $host;
			push @reply, ':'.$config::conf{local}." 311 $src $nick $ident $host * :$gecos";
			push @reply, ':'.$config::conf{local}." 312 $src $nick ".$config::conf{local}.' :'.$config::conf{info};
			foreach my $mode (split(//, $modes)) {
				if ($mode eq 'z') {
					push @reply, ':'.$config::conf{local}." 671 $src $nick :is using a Secure Connection";
				}
				elsif($mode eq 'S') {
					#313 tab ChanServ :is a Network Service
					push @reply, ':'.$config::conf{local}." 313 $src $nick :is a Network Service";
				}
				elsif($mode eq 'B') {
					#335 blah2 TriviaBot :is a Bot on SurrealChat.net
					push @reply, ':'.$config::conf{local}.
						" 335 $src $nick :is a \002Bot\002 on ".$config::conf{network};
				}
			}
		}
		else {
			push @reply, ':'.$config::conf{local}." 401 $src $nick :No such service";
		}
		
	}
	push @reply, ':'.$config::conf{local}." 318 $src $nicklist :End of /WHOIS list.";
	ircsend(@reply);
}

sub parse_tkl ($) {
	my ($in) = @_;
	# This function is intended to accept ALL tkl types,
	# tho maybe not parse all of them in the first version.

	# Discard first token, 'TKL'
	my (undef, $sign, $type, $params) = split(/ /, $in, 4);

	# Yes, TKL types are case sensitive!
	# also be aware (and this applies to the net.pm generator functions too)
	# This implementation may appear naiive, but Unreal assumes that, for a given
	# TKL type, that all parameters are non-null.
	# Thus, if any parameters ARE null, Unreal WILL segfault.
	## Update: this problem may have been fixed since Unreal 3.2.2 or so.
	if ($type eq 'G' or $type eq 'Z' or $type eq 's' or $type eq 'Q') {
		# format is
		# TKL + type ident host setter expiretime settime :reason
		# TKL - type ident host setter
		# for Q, ident is always '*' or 'h' (Services HOLDs)
		if ($sign eq '+') {
			my ($ident, $host, $setter, $expire, $time, $reason) = split(/ /, $params, 6);

			$reason =~ s/^\://;
			return ($type, +1, $ident, $host, $setter, $expire, $time, $reason);
		}
		elsif($sign eq '-') {
			my ($ident, $host, $setter) = split(/ /, $params, 3);
			return ($type, -1, $ident, $host, $setter);
		}
	}
	elsif($type = 'F') {
		# TKL + F cpnNPq b saturn!attitude@netadmin.SCnet.ops 0 1099959668 86400 Possible_mIRC_DNS_exploit :\/dns (\d+\.){3}\d
		# TKL + F u g saturn!attitude@saturn.netadmin.SCnet.ops 0 1102273855 604800 sploogatheunbreakable:_Excessively_offensive_behavior,_ban_evasion. :.*!imleetnig@.*\.dsl\.mindspring\.com
		# TKL - F u Z tabris!northman@tabris.netadmin.SCnet.ops 0 0 :do_not!use@mask
		if ($sign eq '+') {
			my ($target, $action, $setter, $expire, $time, $bantime, $reason, $mask) = split(/ /, $params, 8);
			$mask =~ s/^\://;
			return ($type, +1, $target, $action, $setter, $expire, $time, $bantime, $reason, $mask);
		}
		elsif($sign eq '-') {
			my ($target, $action, $setter, $expire, $time, $mask) = split(/ /, $params, 6);
			$mask =~ s/^\://;
			return ($type, -1, $target, $action, $setter, $mask);
		}
	}
}

# Send Functions

sub ircsendreal {
	foreach my $line (@_) {
		print IRC $line . $nl;
		print "<< $line\n" if $debug;
	}
}

sub ircsend {
	#print "IRCSEND: ready:$ready shutdown:$thread::shutdown\n\n";
	return unless @_;

	if($debug) {
		foreach my $x (@_) {
			print "<< $ircline $x\n";
		}
	}

	unless(thread::alive()) { ircsendreal(@_); return; }
	if($ready) {
		thread::ircsend(@_);
	} else {
		foreach my $x (@_) {
			if($x =~ /^$tkn{NICK}[$tkn]/) {
				unshift @queue, $x;
			} else {
				push @queue, $x;
			}
		}
	}
}

sub ircsendimm {
	thread::ircsend(@_);
}

sub kick($$$$) {
	my ($src, $chan, $target, $reason) = @_;
	$src = $config::conf{local} unless synced();
	ircsend(":$src $tkn{KICK}[$tkn] $chan $target :$reason");
#	thread::ircrecv(":$src $tkn{KICK}[$tkn] $chan $target :$reason");
	callfuncs('KICK', 0, 2, [$src, $chan, $target, $reason]);
}

sub invite($$$) {
	my ($src, $chan, $target) = @_;
	#:SecurityBot INVITE tabris #channel
	ircsend(":$src $tkn{INVITE}[$tkn] $target $chan");
}

sub ping {
#	if(@_ == 1) {
		ircsend(':'.$config::conf{local}.' '.$tkn{PING}[$tkn].' :'.$config::conf{local});
#	} else {
#		ircsend(':'.$_[2].' '.$tkn{PONG}[$tkn].' '.$_[0].' :'.$_[1]);
#	}
}

sub privmsg($$@) {
	my ($src, $dst, @msgs) = @_;

	my @bufs;
	foreach my $buf (@msgs) {
		# 3 spaces, two colons, PRIVMSG=7
		# Length restrictions are for CLIENT Protocol
		# hence the (MASKLEN - (NICKLEN + 1))
		# Technically optimizable if we use $agent{lc $src}'s ident and host
		my $buflen = length($src) + length($dst) + 12 + (MASKLEN - (NICKLEN + 1));
		push @bufs, fmt::wordwrap($buf, (MAXBUFLEN - $buflen));
	}

	# submit a list of messages as a single packet to the server
	ircsend(":$src $tkn{PRIVMSG}[$tkn] $dst :".join("\r\n".":$src $tkn{PRIVMSG}[$tkn] $dst :", @bufs));
}

sub debug(@) {
	my (@msgs) = @_;
	debug_privmsg($config::conf{local}, $config::conf{diag}, @msgs);
	log::write_log('diag', '<'.$config::conf{local}.'>', @msgs);
}

sub debug_nolog(@) {
	my (@msgs) = @_;
	debug_privmsg($config::conf{local}, $config::conf{diag}, @msgs);
}

sub debug_privmsg($$@) {
	my ($src, $dst, @msgs) = @_;

	my @bufs;
	foreach my $buf (@msgs) {
		# 3 spaces, two colons, PRIVMSG=7
		# Length restrictions are for CLIENT Protocol
		# hence the (MASKLEN - (NICKLEN + 1))
		my $buflen = length($src) + length($dst) + 12 + (MASKLEN - (NICKLEN + 1));
		push @bufs, fmt::wordwrap($buf, (MAXBUFLEN - $buflen));
	}

	# submit a list of messages as a single packet to the server
	ircsendimm(":$src $tkn{PRIVMSG}[$tkn] $dst :".join("\r\n".":$src $tkn{PRIVMSG}[$tkn] $dst :", @bufs));
}

sub notice($$@) {
	my ($src, $dst, @msgs) = @_;

	my @bufs;
	foreach my $buf (@msgs) {
		# 3 spaces, two colons, NOTICE=6
		# Length restrictions are for CLIENT Protocol
		# hence the (MASKLEN - (NICKLEN + 1))
		my $buflen = length($src) + length($dst) + 12 + (MASKLEN - (NICKLEN + 1));
		push @bufs, fmt::wordwrap($buf, (MAXBUFLEN - $buflen));
	}

	# submit a list of notices as a single packet to the server
	ircsend(":$src $tkn{NOTICE}[$tkn] $dst :".join("\r\n".":$src $tkn{NOTICE}[$tkn] $dst :", @bufs));
}

sub ctcp($$@) {
	my ($src, $dst, $cmd, @toks) = @_;

	privmsg($src, $dst, "\x01".join(' ', ($cmd, @toks))."\x01");
}

sub ctcp_reply($$@) {
	my ($src, $dst, $cmd, @toks) = @_;

	notice($src, $dst, "\x01".join(' ', ($cmd, @toks))."\x01");
}

sub setumode($$$) {
	my ($src, $dst, $modes) = @_;

	ircsend(":$src $tkn{SVS2MODE}[$tkn] $dst $modes");
	callfuncs('UMODE', 0, undef, [$dst, $modes]);
}

sub setsvsstamp($$$) {
	my ($src, $dst, $stamp) = @_;

	ircsend(":$src $tkn{SVS2MODE}[$tkn] $dst +d $stamp");
	# This function basically set the svsstamp to
	# be the same as the userid. Not all ircd will
	# support this function.
	# We obviously already know the userid, so don't
	# use a callback here.
	#callfuncs('UMODE', 0, undef, [$dst, $modes]);
}

sub setagent_umode($$) {
	my ($src, $modes) = @_;

	ircsend(":$src $tkn{UMODE2}[$tkn] $modes");
}

sub setmode($$$;$) {
	my ($src, $dst, $modes, $parms) = @_;
	$src = $config::conf{local} unless synced();

	callfuncs('MODE', undef, 1, [$src, $dst, $modes, $parms]);
	
	print "$ircline -- setmode($src, $dst, $modes, $parms)\n" if $debug;
	my $prev = $defer_mode{"$src $dst"}[-1];

	if(defined($prev)) {
		my ($oldmodes, $oldparms) = split(/ /, $prev, 2);
		
		# 12 modes per line
		if((length($oldmodes.$modes) - @{[($oldmodes.$modes) =~ /[+-]/g]}) <= 12 and length($src.$dst.$parms.$oldparms) < 400) {
			$defer_mode{"$src $dst"}[-1] = modes::merge(
				$prev, "$modes $parms", ($dst =~ /^#/ ? 1 : 0));
			print $defer_mode{"$src $dst"}[-1], " *** \n" if $debug;
			
			return;
		}
	}
	
	push @{$defer_mode{"$src $dst"}}, "$modes $parms";
}

sub flushmodes {
	my @k = keys(%defer_mode); my @v = values(%defer_mode);
	
	for(my $i; $i<@k; $i++) {
		my ($src, $dst) = split(/ /, $k[$i]);
		my @m = @{$v[$i]};
		foreach my $m (@m) {
			my ($modes, $parms) = split(/ /, $m, 2);

			setmode_real($src, $dst, $modes, $parms);
		}
	}

	%defer_mode = ();
}

sub setmode_real($$$;$) {
	my ($src, $dst, $modes, $parms) = @_;

	print "$ircline -- setmode_real($src, $dst, $modes, $parms)\n" if $debug;
	ircsend(":$src ".$tkn{MODE}[$tkn]." $dst $modes".($parms?" $parms":''));
}

sub settopic($$$$$) {
	my ($src, $chan, $setter, $time, $topic) = @_;
	
	ircsend(":$src ".$tkn{TOPIC}[$tkn]." $chan $setter $time :$topic");
	callfuncs('TOPIC', undef, undef, [$src, $chan, $setter, $time, $topic]);
}

sub wallops ($$) {
	my ($src, $message) = @_;
	ircsend(":$src $tkn{WALLOPS}[$tkn] :$message");
}

sub globops ($$) {
	my ($src, $message) = @_;
	ircsend(":$src $tkn{GLOBOPS}[$tkn] :$message");
}

sub kline ($$$$$) {
        my ($setter, $ident, $host, $expiry, $reason) = @_;
	$setter=$config::conf{local} unless defined($setter);
	$ident = '*' unless defined($ident);


	#foreach my $ex (@except) { return 1 if $mask =~ /\Q$ex\E/i; }
	
	#my $line = "GLINE $mask $time :$reason";
	# you need to use TKL for this. GLINE is a user command
	# TKL is a server command.	
        # format is
        # TKL +/- type ident host setter expiretime settime :reason
#:nascent.surrealchat.net TKL + G * *.testing.only tabris!northman@netadmin.SCnet.ops 1089168439 1089168434 :This is just a test.
        my $line = "TKL + G $ident $host $setter ".($expiry + time()).' '.time()." :$reason";

	#notice($nick, $conf{"Debugger"}, $line) if $conf{"Debug"};
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub unkline ($$$) {
	my ($setter, $ident, $host) = @_;
	# TKL - G ident host setter
# TKL - G ident *.test.dom tabris!northman@netadmin.SCnet.ops
	my $line = "TKL - G $ident $host $setter";
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub zline ($$$$) {
        my ($setter, $host, $expiry, $reason) = @_;
	$setter=$config::conf{local} unless defined($setter);

	#foreach my $ex (@except) { return 1 if $mask =~ /\Q$ex\E/i; }
	
        # format is
        # TKL +/- type ident host setter expiretime settime :reason
        my $line = "TKL + Z * $host $setter ".($expiry + time).' '.time." :$reason";
	#notice($nick, $conf{"Debugger"}, $line) if $conf{"Debug"};
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub unzline ($$) {
	my ($setter, $host) = @_;
	# TKL - G ident host setter
# TKL - G ident *.test.dom tabris!northman@netadmin.SCnet.ops
	my $line = "TKL - Z * $host $setter";
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub spamfilter($$$$$$$) {
# Note the hardcoded zero (0).
# Looks like theoretically one can have expirable spamfilters.
# This is untested however.
	my ($sign, $tkl_target, $tkl_action, $setter, $bantime, $reason, $regex) = @_;
	my $tkl = "TKL ".($sign ? '+' : '-' )." F $tkl_target $tkl_action $setter 0 ".time()." $bantime $reason :$regex";
	ircsend($tkl);
	callfuncs('TKL', undef, undef, [parse_tkl($tkl)]);
}

sub irckill($$$) {
	my ($src, $target, $reason) = @_;
	$src = $config::conf{local} unless synced();
	# This is a simple way to do it, that _could_ be defeated
	# with enough users getting killed at once.
	# The alternative would require a timer to expire the old entries.
	{
		lock @userkill;
		return undef if (time() == $userkill[1] and $target eq $userkill[0]);
		@userkill = ($target, time());
		ircsendimm(":$src ".$tkn{KILL}[$tkn]." $target :$src ($reason)");
	}
	# Don't want the lock held while waiting for all of the callbacks.
	# Always minimize lock time
	callfuncs('KILL', 0, 1, [$src, $target, $src, $reason]);
	return 1;
}

sub svssno($$$) {
    my ($src, $target, $snomasks) = @_;
    $src=$config::conf{local} unless defined($src);
    # TODO:
    # None, this doesn't affect us.

    # SVSSNO is not in tokens.txt nor msg.h
    ircsend(":$src ".'SVS2SNO'." $target $snomasks ".time);
}

sub svsnick($$$) {
    my ($src, $oldnick, $newnick) = @_;
    $src=$config::conf{local} unless defined($src);
    # note: we will get a NICK cmd back after a 
    # successful nick change.
    # warning, if misused, this can KILL the user
    # with a collision
    
#    ircsend(":$src ".$tkn{SVSNICK}[$tkn]." $oldnick $newnick ".time);
    ircsend($tkn{SVSNICK}[$tkn]." $oldnick $newnick :".time);
}

sub svsnoop($$$) {
    my ($targetserver, $bool, $src) = @_;
    $src = $config::conf{local} unless defined($src);
    if ($bool > 0) { $bool = '+'; } else { $bool = '-'; }
#this is SVS NO-OP not SVS SNOOP
    ircsend(":$config::conf{local} $tkn{SVSNOOP}[$tkn] $targetserver $bool");
}

sub svswatch ($$@) {
# Changes the WATCH list of a user.
# Syntax: SVSWATCH <nick> :<watch parameters>
# Example: SVSWATCH Blah :+Blih!*@* -Bluh!*@* +Bleh!*@*.com
# *** We do not track this info nor care.
	my ($src, $target, @watchlist) = @_;
	my $base_str = ":$src ".$tkn{SVSWATCH}[$tkn]." $target :";
	my $send_str = $base_str;
	while (@watchlist) {
		my $watch = shift @watchlist;
		if (length("$send_str $watch") > MAXBUFLEN) {
			ircsend($send_str);
			$send_str = $base_str;
		}
		$send_str = "$send_str $watch";
	}
	ircsend($send_str);
}

sub svssilence ($$@) {
# Changes the SILENCE list of a user.
# Syntax: SVSSILENCE <nick> :<silence parameters>
# Example: SVSSILENCE Blah :+Blih!*@* -Bluh!*@* +Bleh!*@*.com
# *** We do not track this info nor care.
	my ($src, $target, @silencelist) = @_;
	my $base_str = ":$src ".$tkn{SVSSILENCE}[$tkn]." $target :";
	my $send_str = $base_str;
	while (@silencelist) {
		my $silence = shift @silencelist;
		if (length("$send_str $silence") > MAXBUFLEN) {
			ircsend($send_str);
			$send_str = $base_str;
		}
		$send_str = "$send_str $silence";
	}
	ircsend($send_str);
}

sub svso($$$) {
# Gives nick Operflags like the ones in O:lines.
# SVSO <nick> <+operflags> (Adds the Operflags)
# SVSO <nick> - (Removes all O:Line flags)
# Example: SVSO SomeNick +bBkK
# *** We do not track this info nor care.
# *** We will see any umode changes later.
# *** this cmd does not change any umodes!

    my ($src, $target, $oflags) = @_;
    $src = $config::conf{local} unless defined($src);
    ircsend(":$src $tkn{SVSO}[$tkn] $target $oflags");

}

sub swhois($$$) {
# *** We do not track this info nor care.
    my ($src, $target, $swhois) = @_;
    $src = $config::conf{local} unless defined($src);
    ircsend(":$src $tkn{SWHOIS}[$tkn] $target :$swhois");
}

sub svsjoin($$@) {
    my ($src, $target, @chans) = @_;
    # a note. a JOIN is returned back to us on success
    # so no need to process this command.
    # similar for svspart.
    ircsend(($src?":$src":'')." $tkn{SVSJOIN}[$tkn] $target ".join(',', @chans));
}

sub svspart($$$@) {
    my ($src, $target, $reason, @chans) = @_;
    ircsend(($src ? ":$src" : '')." $tkn{SVSPART}[$tkn] $target ".join(',', @chans).
    	($reason ? " :$reason" : ''));
}

sub sqline ($;$) {
# we need to sqline most/all of our agents.
# tho whether we want to put it in agent_connect
# or leave it to the module to call it...
	my ($nickmask, $reason) = @_;
	#ircsend("$tkn{S1QLINE}[$tkn] $nickmask".($reason?" :$reason":''));
	qline($nickmask, 0, $reason);
}

sub svshold($$$) {
# Not all IRCd will support this command, as such the calling module must check the IRCd capabilities first.
	my ($nickmask, $expiry, $reason) = @_;
# TKL version - Allows timed qlines.
# TKL + Q * test services.SC.net 0 1092179497 :test
	my $line = 'TKL + Q H '.$nickmask.' '.$config::conf{local}.' '.($expiry ? $expiry+time() : 0).' '.time().' :'.$reason;
	ircsend($line);

	# at startup we send these too early,
	# before the handlers are initialized
	# so they may be lost.
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub svsunhold($) {
	my ($nickmask) = @_;
# TKL version
# TKL - Q * test services.SC.net
	my $line = 'TKL - Q H '.$nickmask.' '.$config::conf{local};
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub qline($$$) {
	my ($nickmask, $expiry, $reason) = @_;
# TKL version - Allows timed qlines.
# TKL + Q * test services.SC.net 0 1092179497 :test
	my $line = 'TKL + Q * '.$nickmask.' '.$config::conf{local}.' '.($expiry ? $expiry+time() : 0).' '.time().' :'.$reason;
	ircsend($line);

	# at startup we send these too early,
	# before the handlers are initialized
	# so they may be lost.
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub unsqline ($) {
# we need to sqline most/all of our agents.
# tho whether we want to put it in agent_connect
# or leave it to the module to call it...
	my ($nickmask) = @_;
	unqline($nickmask);
}

sub unqline($) {
	my ($nickmask) = @_;
# TKL version
# TKL - Q * test services.SC.net
	my $line = 'TKL - Q * '.$nickmask.' '.$config::conf{local};
	ircsend($line);
	callfuncs('TKL', undef, undef, [parse_tkl($line)]);
}

sub svskill($$$) {
	my ($src, $target, $reason) = @_;
	# SVSKILL requires a src, it will NOT work w/o one.
	# not sure if it'll accept a servername or not.
	# consider defaulting to ServServ
	die('svskill called w/o $src') unless $src;
	ircsend(':'.$src.' '.$tkn{SVSKILL}[$tkn].' '.$target.' :'.$reason);
	callfuncs('QUIT', 0, undef, [$target, $reason]);
}

sub version($) {
	my ($src) = @_;
	ircsend(":$config::conf{local} 351 $src $main::progname ver $main::version $config::conf{local} ".
		$main::extraversion);
}

sub userhost($) {
	my ($target) = @_;
	ircsend($tkn{USERHOST}[$tkn]." $target");
}

sub userip($) {
	my ($target) = @_;
	die "We're not supposed to use USERIP anymore!" if $debug and $nickip;
	ircsend(":$main::rsnick USERIP $target");
}

sub chghost($$$) {
	my ($src, $target, $vhost) = @_;
	ircsend(($src?":$src ":'').$tkn{CHGHOST}[$tkn]." $target $vhost");
        callfuncs('CHGHOST', 0, 1, [$src, $target, $vhost]);
}

sub chgident($$$) {
	my ($src, $target, $ident) = @_;
	ircsend(($src?":$src ":'').$tkn{CHGIDENT}[$tkn]." $target $ident");
        callfuncs('CHGIDENT', 0, 1, [$src, $target, $ident]);
}

sub jupe_server($$) {
	my ($server, $reason) = @_;

	# :nascent.surrealchat.net SERVER wyvern.surrealchat.net 2 :SurrealChat
	ircsend(':'.$config::conf{local}.' '.$tkn{SERVER}[$tkn]." $server 2 :$reason");

	# We don't need a callback as long as we don't care about spurious SQUITs
	# Consider set_server_state($server, -1) for a special value for jupes
}

sub rehash_all_servers(;$) {
	my ($type) = @_;

	# Validate the type before passing it along.
	# Very IRCd specific! May be version specific.
	$type = undef() if(defined($type) && !($type =~ /^\-(motd|botmotd|opermotd|garbage)$/i));

	foreach my $server (get_online_servers()) {
		ircsend(':'.$main::rsnick.' '.$tkn{REHASH}[$tkn].' '.$server.(defined($type) ? ' '.$type : '') );
	}
}

1;
